;=============================================================================
; Insight, real-mode debugger for MS DOS / PC DOS / FreeDOS.
; Copyright (c) Victor M. Gamayunov, Sergey Pimenov, 1993, 96, 97, 2002.
; Modifications by Oleg O. Chukaev (2006 - 2009).
;-----------------------------------------------------------------------------
; keyboard.inc
; Procedures for working with keyboard.
;-----------------------------------------------------------------------------
; This program is free software; you can redistribute it and/or
; modify it under the terms of the GNU General Public License
; as published by the Free Software Foundation; either version 2
; of the License, or (at your option) any later version.
;
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
; 
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
; 02110-1301 USA
;=============================================================================


;=============================================================================
; Changelog
;-----------------------------------------------------------------------------
; 2007-02-10	0leg	Added comments to all procedures. All variables moved
;			to act[u]data.inc
;=============================================================================


;=============================================================================
; key_loop
;-----------------------------------------------------------------------------
; Reads keystrokes and calls corresponding procedures in loop.
; On end of table key_loop calls corresponding proc, if it is not 0.
; If it is 0 or if it return CY, [test_key] called. If [test_key]
; return NC, key_loop terminates, otherwise beep called.
; In:	[keys] -> structure: key-code: word, proc: word,
;		end of table: key-code == 0, exit: proc == 0
; Out:	AX -- key codes
; Modf:	AX, BX, CX, DX, SI, DI, BP
; Call:	flush_key, read_key, [test_key], [key_loop_sub], beep
; Use:	keys, key_loop_sub, cur_test_code, test_key
;
key_loop:

@@loop:
		push	ax
		push	bx
		push	si
		mov	si,[keys]
		call	flush_key
		call	read_key
		xchg	ax,bx
@@next:
		lodsw
		or	ax,ax			;End of table?
		jz	@@not_found		;Yes
		cmp	ax,bx
		je	@@found
		lodsw
		jmp	@@next

@@found:
		lodsw
		or	ax,ax			;Exit?
		jz	@@quit			;Yes
		mov	[key_loop_sub],ax
		pop	si
		pop	bx
		pop	ax
		call	word [key_loop_sub]	;Call procedure
		jmp	@@loop

@@not_found:
		mov	[cur_test_code],bx	;Key codes
		lodsw
		mov	[key_loop_sub],ax	;Procedure for any key
		or	ax,ax			;Exit?
		xchg	ax,bx
		pop	si
		pop	bx
		jz	@@test_key		;Yes
		call	word [key_loop_sub]	;AX == codes on entry
		jc	@@test_key
		pop	ax
		jmp	@@loop			;Next...

@@test_key:
		pop	ax
		call	word [test_key]
		jnc	@@quit_1
		call	beep
		jmp	@@loop
@@quit:
		mov	ax,bx			;Key codes
		pop	si
		pop	bx
		inc	sp
		inc	sp
@@quit_1:
		ret
;=============================================================================
; def_test_key
;-----------------------------------------------------------------------------
; Default key testing procedure.
; In:	---
; Out:	CY
; Modf:	---
; Call:	---
; Use:	---
;
def_test_key:
		stc
		ret
;=============================================================================
; read_key
;-----------------------------------------------------------------------------
; Reads keystroke.
; In:	---
; Out:	AX -- ASCII/scan codes
; Modf:	AX
; Call:	---
; Use:	---
;
read_key:
	xor ax, ax
	xchg ax, word [cs:read_key_insert]
	test ax, ax
	jnz .ret
		push	bx
		xor	bx,bx
	jmp @F
@@next:
%if _TERMINAL
	call terminal_receive_to_buffer
%endif
	call idle			; idle while waiting
@@:

	; The 16.10 Enhanced Keyboard Support is done in a way
	;  that lack of support might be detected here, so that
	;  the user isn't left unable to control the application
	;  if they mistakenly enable the option to use 16.10.
		mov	ah,1
	rol byte [option_use_16_10], 1
	jnc @F
	mov ah, 11h
	test ax, ax			; (NZ in case 16.11 unsupported)
@@:
		int	16h
		jnz	@@get		; got a keypress (or unsupported) -->
		mov	ah,2
		int	16h
		test	al,8
		jnz	@@alt
		cmp	bl,2
		je	@@ret
		mov	bl,1
		jmp	@@next
@@ret:
		mov	ax,kbAltAlone
		jmp	@@quit
@@get:
		mov	ah,0
	rol byte [option_use_16_10], 1
	jnc @F
	mov ax, 1000h
@@:
		int	16h		; get and eat keypress
	rol byte [option_use_16_10], 1
	jnc @FF				; not using 16.10 -->
	cmp ax, 1000h			; Alt+Q or 16.10 not supported ?
	jne @F				; returned something other than Alt+Q -->
	inc byte [altq_16_10_unsupported_counter]
	jnz @FF				; not yet wrapped -->
	not byte [option_use_16_10]	; reset if it appears unsupported

@@:				; reset counter if got something other than Alt+Q
	mov byte [altq_16_10_unsupported_counter], 0
@@:
@@quit:
		pop	bx
.ret:
%if _EINSICHT
	mov byte [einsicht_send_beeped], 0
%endif
		ret
@@alt:
		or	bx,bx
		jz	@@next
		mov	bl,2
		jmp	@@next

%if _TERMINAL
terminal_receive_to_buffer:
	call pushr
	push cs
	pop ds
	rol byte [terminal_connected], 1
	jnc .shownix
.loop:
	push cs
	pop es
		call IRQ_READBUF        ; ***** fetch data from input buffer
	test ax, ax
	js .shownix	; only show properly received chars
			; (do not show the -1 we get if no chars received :-))
%if _EINSICHT
	mov di, word [einsicht_receive_buffer.end]
	add di, einsicht_receive_buffer

	cmp al, 10
	jne .not_einsicht_lf

	mov byte [di], 0
	cmp di, einsicht_receive_buffer
	jbe @F
	cmp byte [di - 1], 13
	jne @F
	mov byte [di - 1], 0
@@:
..@continue_after_more:
	mov si, einsicht_receive_buffer
	mov bx, einsicht_keywords
@@:
	mov di, word [bx]
	test di, di
	jz .einsicht_lf_done_no_keyword
	mov cx, word [di]
	scasw

	add bx, 4
	push si
	repe cmpsb
	pop si
	jne @B
	call near [bx - 2]

.einsicht_lf_done:
	mov al, 10
	and word [einsicht_receive_buffer.end], 0
	jmp @F


.einsicht_lf_done_no_keyword:
	call handle_einsicht_data_dump
	jmp .einsicht_lf_done


.not_einsicht_lf:
	mov byte [einsicht_at_prompt], 0

	cmp di, einsicht_receive_buffer.end - 1
	jae @F
	mov byte [di], al
	inc word [einsicht_receive_buffer.end]

	cmp al, ']'
	jne @F
	cmp di, einsicht_receive_buffer - 1 + 6
	jne @F
	mov si, einsicht_receive_buffer
	mov di, einsicht_keywords.more + 2
	mov cx, 6
	repe cmpsb
	jne @F
	rol byte [terminal_active], 1
	jc @F
	mov al, 32
	call SENDME
	mov al, ']'
	jmp .notprompt

@@:

	cmp al, '-'
	je .check_prompt
	cmp al, '#'
	je .check_prompt
	cmp al, 32
	je .check_keep
	jmp .notprompt


.check_keep:
	cmp byte [di - 1], '='
	jne .notprompt

	push ax
	mov ax, di
	inc ax
	mov si, einsicht_receive_buffer
	mov bx, einsicht_keywords.keepprompts
	sub ax, si

%if 1
@@:
	mov di, word [bx]
	test di, di
	jz .notprompt_pop
	inc bx
	inc bx

	mov cx, word [di]
	scasw
	jne @B
	push si
	repe cmpsb
	pop si
	jne @B
%endif

	mov di, einsicht_send_buffer
	mov ax, "KE"
	stosw
	mov ax, "EP"
	stosw
	call einsicht_send_command.no_prompt_check
	jmp .notprompt_pop


.check_prompt:
	push ax
	mov ax, di
	inc ax
	mov si, einsicht_receive_buffer
	mov bx, einsicht_keywords.prompts
	sub ax, si
@@:
	mov di, word [bx]
	test di, di
	jz .notprompt_pop
	add bx, 4

	mov cx, word [di]
	scasw
	jne @B
	push si
	repe cmpsb
	pop si
	jne @B

	mov byte [einsicht_at_prompt], -1
.notprompt_pop:
	pop ax
.notprompt:
%endif
	mov di, word [terminal_buffer.end]
	mov es, word [terminal_buffer]
	mov ah, byte [es:di]
	stosb
	inc word [terminal_buffer.end]
	 mov bx, word [terminal_buffer.draw]
	cmp di, word [terminal_buffer.start]
	jne .leave
	push ax
	mov si, word [terminal_buffer.start]
	call .inc_bx_if_equal
	inc si
	xchg ah, al
@@:
	cmp al, 10
	je @F
	 call .inc_bx_if_equal
	es lodsb
	jmp @B

@@:
	cmp word [terminal_buffer.draw], bx
	je @F
	dec word [terminal_buffer.lf_amount]
@@:
	 mov word [terminal_buffer.draw], bx
	mov word [terminal_buffer.start], si
	pop ax
.leave:

	cmp al, 10
	jne .not_write_lf
	cmp word [terminal_buffer.lf_amount], MAX_SCR_ROWS
	jbe .increment_lf_amount
	xchg bx, si
@@:
	cmp si, word [terminal_buffer.end]
	je @F				; (unlikely, we had lf_amount > rows)
	es lodsb
	cmp al, 10
	jne @B
@@:
	mov al, 10
	mov word [terminal_buffer.draw], si
	jmp .not_write_lf

.increment_lf_amount:
	inc word [terminal_buffer.lf_amount]

.not_write_lf:
	rol byte [terminal_active], 1
	jnc .loop
	mov es, [video_seg]
	call SHOW_RX
	jmp .loop
.shownix:
	call popr
	retn

.inc_bx_if_equal:
	cmp si, bx
	jne @F
	inc bx
@@:
	retn
%endif


%if _EINSICHT
einsicht_mem_readonly:
	mov byte [einsicht_readonly_detected], 0FFh
	retn

einsicht_cut_more:
	mov di, si
	add si, 1 + 4 + 1 + 1 + 6 + 1
@@:
	lodsb
	stosb
	test al, al
	jnz @B
	pop ax
	jmp ..@continue_after_more

einsicht_prompt_received:
	mov byte [terminal_prompt_received], -1
	retn

einsicht_parse_r_16_1:
	mov byte [einsicht_reg32_unknown], -1
	jmp einsicht_parse_register

einsicht_parse_r_16_2_or_r_32_3: equ einsicht_parse_register
einsicht_parse_r_40_2: equ einsicht_parse_register
einsicht_parse_r_40_3: equ einsicht_parse_register

einsicht_parse_r_32_1:
	mov byte [einsicht_reg32_unknown], 0
	jmp einsicht_parse_register

einsicht_parse_r_32_2: equ einsicht_parse_register

einsicht_parse_fl:
	add si, 3
	call get_hex_value
	jc .ret
	mov word [cpu_regs + REG_FL], bx
.ret:
	jmp update_registers

einsicht_parse_efl:
	add si, 4
	call get_dword_hex_value
	jc .ret
	mov word [cpu_regs + REG_FL], dx
	mov word [cpu_regs + REG_FL + 2], bx
.ret:
	jmp update_registers


@@:
	dec si
einsicht_parse_register:
.:
	lodsw			; ax = name if 16-bit
	cmp al, 32
	je @B
	cmp al, 9
	je @B
	cmp al, 13		; treat as a blank
	je @B
	cmp al, 10		; likewise
	je @B
	cmp al, 0		; EOL ?
	je .ret
	push ax			; save name
	push si
	cmp al, 'E'		; can be extended (32-bit) register ?
	jne @F			; no -->
	dec si
	lodsw			; get corresponding 16-bit name if 32-bit
	cmp ah, 'S'		; ExS invalid
	je @F			; check for 16-bit reg -->
	mov di, rreg
	mov cx, 16
	repne scasw		; valid ?
	jne @F			; no, try for 16-bit reg (could be ES) -->
	pop ax			; discard
	pop ax			; discard
	lodsb			; get equals sign
	cmp al, '='		; is it ?
	jne .ret		; no -->
	call einsicht_register_match_to_variable
				; di -> our variable
	call get_dword_hex_value
				; get 32-bit value
	jc .ret			; if no hexit or too large -->
	mov word [di], dx
	mov word [di + 2], bx	; store this away
	jmp .			; back for more

@@:
	pop si			; return pointer
	pop ax			; get candidate 16-bit reg name
	mov di, rreg
	mov cx, 16
	repne scasw		; valid 16-bit reg ?
	jne .ret		; no -->
	lodsb			; get equals sign
	cmp al, '='		; is it ?
	jne .ret		; no -->
	call einsicht_register_match_to_variable
	call get_hex_value	; get 16-bit value
	jc .ret			; if no hexit or too large -->
	mov word [di], bx	; store this away
	jmp .			; back for more

.ret:
	jmp update_registers


einsicht_register_match_to_variable:
				; di -> behind rreg matching name
	sub di, rreg + 2	; di = index times 2
	shl di, 1		; = index times 4
	add di, cpu_regs	; -> variable
	retn
%endif

;=============================================================================
; in_key
;-----------------------------------------------------------------------------
; Checks if key pressed.
; In:	---
; Out:	ZR if no key pressed
;	NZ if key pressed:
;		AX -- ASCII/scan codes
; Modf:	AX
; Call:	---
; Use:	---
;
%if 0
in_key:

		mov	ah,1
		int	16h
		ret
%endif
;=============================================================================
; flush_key
;-----------------------------------------------------------------------------
; Flushes keyboard buffer.
; In:	---
; Out:	---
; Modf:	AX
; Call:	---
; Use:	---
;
flush_key:
@@next:
		mov	ah,1
		int	16h
		jz	@@quit
		mov	ah,0
		int	16h
		jmp	@@next
@@quit:
		ret
;=============================================================================
; check_shift
;-----------------------------------------------------------------------------
; Checks if any shift key is pressed.
; In:	---
; Out:	ZR if shift key is not pressed
;	NZ if shift key is pressed
; Modf:	---
; Call:	---
; Use:	---
;
check_shift:
		push	ax
		push	ds
		xor	ax,ax
		mov	ds,ax
		test	byte [417h],11b
		pop	ds
		pop	ax
		ret


		; INP:	-
		; OUT:	-
		; CHG:	ax
		; STT:	ds = ss = debugger segment/selector
		;
		; Idle system, using 2F.1680 or sti \ hlt.
idle:
	sti
		; This instruction cures a failure in the loop
		;  in run_push_button. Although dosemu2 dosdebug
		;  does say IF is on, without this instruction
		;  no time tick interrupts arrive anymore.
		; So it ended up in an infinite loop.

		; Test case was: run insight.com /E /A lclock.com,
		;  run lClock with Alt+F9, dismiss return code popup,
		;  open Configuration with F10, O, N, press Alt+C.
		; Cancel button will be animated as pressed but then
		;  the wait loop in run_push_button loops infinitely.
	nop			; help debugging

	; testopt [options3], opt3_no_idle_2F
	; jnz .hlt
	rol byte [option_1680], 1
	jnc .hlt
 ; %if _GUARD_86M_INT2F
 %if 1
	push es
	xor ax, ax
	mov es, ax		; (only used in 86 Mode)
	mov ax, [es:2Fh * 4]
	cmp ax, -1
	je @F			; --> (ZR)
	or ax, [es:2Fh * 4 + 2]
@@:
	pop es
	jz @FF
@@:
 %endif
	mov ax, 1680h
	int 2Fh			; release timeslice in multitasker
	test al, al
	jz .return		; done idling -->
@@:
.hlt:
	; testopt [options], nohlt
	; jnz .return
	rol byte [option_hlt], 1
	jnc .return
	sti
	hlt			; else idle by hlt
	nop
.return:
	retn

;=============================================================================
; E0F
;=============================================================================
